use std::{error::Error, io::{self, Read, Write}};

use protobuf::{CodedInputStream, CodedOutputStream, Message, rt::compute_raw_varint64_size};

use proto::sass::{inbound_message::{CanonicalizeResponse, CompileRequest, FileImportResponse, FunctionCallResponse, ImportResponse, VersionRequest}, InboundMessage, OutboundMessage};

#[derive(Clone, Debug)]
pub enum MessageType {
    CompileRequest(CompileRequest),
    CanonicalizeResponse(CanonicalizeResponse),
    FileImportResponse(FileImportResponse),
    FunctionCallResponse(FunctionCallResponse),
    ImportResponse(ImportResponse),
    VersionRequest(VersionRequest),
}

impl MessageType {
    /// Parses an `InboundMessage` into a `MessageType`.
    pub fn parse(mut message: InboundMessage) -> Result<Self, Box<dyn Error>> {
        match () {
            _ if message.has_compile_request() => Ok(MessageType::CompileRequest(message.take_compile_request())),
            _ if message.has_canonicalize_response() => Ok(MessageType::CanonicalizeResponse(message.take_canonicalize_response())),
            _ if message.has_file_import_response() => Ok(MessageType::FileImportResponse(message.take_file_import_response())),
            _ if message.has_function_call_response() => Ok(MessageType::FunctionCallResponse(message.take_function_call_response())),
            _ if message.has_import_response() => Ok(MessageType::ImportResponse(message.take_import_response())),
            _ if message.has_version_request() => Ok(MessageType::VersionRequest(message.take_version_request())),
            _ => Err("Unknown message type".into()),
        }
    }
}

/// Reads a message from stdin and returns a tuple of the message type and the message ID.
pub fn read_message() -> Result<(MessageType, u64), Box<dyn Error>> {
    let stdin = io::stdin();
    let mut stdin_handle = stdin.lock();
    let mut input_stream: CodedInputStream<'_> = CodedInputStream::new(&mut stdin_handle);
    let length = input_stream.read_raw_varint64()?;
    let id = input_stream.read_raw_varint64()?;
    let mut message_buf = vec![0; length as usize - (compute_raw_varint64_size(id as u64) as usize)];
    let mut bytes_read = 0;
    while bytes_read < message_buf.len() {
        let chunk = match input_stream.read(&mut message_buf[bytes_read..]) {
            Ok(chunk) => chunk,
            Err(e) => {
                eprintln!("Error reading from stdin: {}", e);
                std::process::exit(1);
            }
        };
        if chunk == 0 {
            return Err(Box::new(io::Error::new(
                io::ErrorKind::Other,
                "failed to read whole buffer",
            )));
        }
        bytes_read += chunk;
    }
    let raw_message: InboundMessage = InboundMessage::parse_from_bytes(&message_buf)?;
    let message = MessageType::parse(raw_message)?;
    log::trace!("Message read from stdin");
    log::debug!("Message Type: {:#?}", &message);
    Ok((message, id))
}

/// Writes a message to stdout.
pub fn write_message(message: OutboundMessage, id: u64) -> Result<(), Box<dyn Error>> {
    let mut stdout = io::stdout();
    let mut stdout_handle = stdout.lock();
    let mut output_stream: CodedOutputStream<'_> = CodedOutputStream::new(&mut stdout_handle);
    let length = message.compute_size() + compute_raw_varint64_size(id);
    output_stream.write_raw_varint64(length)?;
    output_stream.write_raw_varint64(id)?;
    message.write_to_writer(&mut output_stream)?;
    output_stream.flush()?;
    stdout.flush()?;
    log::trace!("Message written to stdout");
    Ok(())
}