use std::{
    error::Error,
    path::{Path, PathBuf},
    sync::{Arc, Mutex},
};

use grass_compiler::{
    sass_value::{ArgumentResult, Value},
    Builtin, Result as SassResult, Visitor,
};

use paste::item;

use grass::Fs;

use proto::sass::{
    inbound_message::CompileRequest,
    outbound_message::{
        CompileResponse,
        compile_response::{CompileFailure, CompileSuccess},
    },
    OutboundMessage,
};

use crate::{
    buf_io::{MessageType, read_message, write_message},
    importer_registry::ImporterRegistry,
};

#[derive(Debug)]
struct ImporterFs {
    registry: Arc<Mutex<ImporterRegistry>>,
}

impl Fs for ImporterFs {
    fn is_file(&self, _: &Path) -> bool {
        true
    }

    fn is_dir(&self, _: &Path) -> bool {
        true
    }

    fn read(&self, path: &Path) -> std::io::Result<Vec<u8>> {
        self.registry
            .lock()
            .expect("Failed to lock registry")
            .load(path)
    }

    fn canonicalize(&self, path: &Path) -> std::io::Result<PathBuf> {
        self.registry
            .lock()
            .expect("Failed to lock registry")
            .canonicalize(path)
    }
}

pub fn compile_request(mut request: CompileRequest, id: u64) -> Result<(), Box<dyn Error>> {
    use grass_compiler::sass_value::QuoteKind;
    use proto::sass::outbound_message::FunctionCallRequest;

    log::info!("Starting Compile");

    let mut compile_response = CompileResponse::new();
    let mut options = grass::Options::default()
        .style(grass::OutputStyle::Expanded)
        .allows_charset(false)
        .input_syntax(grass::InputSyntax::Scss);

    let mut importer_registry = ImporterRegistry::new(id, "rust-sass".to_string());

    log::trace!("{:?}", request.global_functions);

    match request.has_string() {
        true => {
            let input = request.take_string();
            let input_string = input.source;
            let url = input.url;
            let parent_path = PathBuf::from(url)
                .parent()
                .expect("Could not get parent path")
                .to_str()
                .expect("Could not convert parent path to string")
                .to_string();
            importer_registry.set_parent_path(parent_path);

            if let Some(importer) = input.importer.0 {
                importer_registry.register(importer);
            } else {
                log::warn!("No Importer");
            }

            for importer in request.importers {
                importer_registry.register(Box::new(importer));
            }

            let importer_registry_mutex: Arc<Mutex<ImporterRegistry>> = Arc::new(Mutex::new(importer_registry));

            macro_rules! define_function {
                ($name:expr, $data:expr, $id:expr) => {
                    item! {
                        fn [<$name>] (mut args: ArgumentResult, _visitor: &mut Visitor<'_>) -> SassResult<Value> {
                            log::trace!("function called");
                            log::trace!("args {:?}", args);

                            let mut function_call_request = FunctionCallRequest::default();
                            let msgid: u32 = rand::random();
                            function_call_request.set_name(String::from($data));
                            function_call_request.id = msgid;
                            let mut arguments: Vec<proto::sass::Value> = Vec::new();
                            let mut n = 0;
                            loop {
                                match args.get_positional(n) {
                                    Some(arg) => {
                                        let span = arg.span;
                                        let str = match arg.node.unquote().to_css_string(span, false)
                                        {
                                            Ok(str) => str,
                                            Err(err) => {
                                                log::error!("Failed to inspect arg");
                                                log::error!("{}", err);
                                                continue;
                                            }
                                        };
                                        log::info!("arg: {}", str);
                                        let string = proto::sass::value::String { text: str.to_string(), quoted: true, ..Default::default() };
                                        let mut value = proto::sass::Value::new();
                                        value.set_string(string);
                                        arguments.push(value);
                                    },
                                    None => break,
                                };
                                n+=1;
                            }
                            function_call_request.arguments = arguments;
                            let mut outbound_message = OutboundMessage::new();
                            outbound_message.set_function_call_request(function_call_request);
                            write_message(outbound_message,$id).expect("Failed to write message");
                            loop {
                                let response = read_message().expect("Failed to read message");
                                log::trace!("event: {:?}", &response.0);
                                if let MessageType::FunctionCallResponse(resp) = response.0 {
                                    match resp.result.unwrap() {
                                        proto::sass::inbound_message::function_call_response::Result::Error(e) => {
                                            log::error!("Error: {:?}", e);
                                            return Ok(Value::Null)
                                        }
                                        proto::sass::inbound_message::function_call_response::Result::Success(v) => {
                                            log::info!("Success: {:?}", v);
                                            match v.value.unwrap() {
                                                proto::sass::value::Value::String(s) => {
                                                    let quote_kind = match s.quoted {
                                                        true => QuoteKind::Quoted,
                                                        false => QuoteKind::None,
                                                    };
                                                    return Ok(Value::String(s.text, quote_kind))
                                                }
                                                _ => {
                                                    unimplemented!("Not implemented")
                                                }
                                            }

                                        }
                                        _ => {}
                                    }
                                    break;
                                } else {
                                    log::trace!("got event with wrong response type")
                                }
                                log::trace!("c");

                            }
                            log::trace!("d");
                            Ok(Value::Null)
                        }
                        options = options.add_custom_fn($name, Builtin::new([<$name>]));
                    }
                };
            }

            define_function!("asset-path", "asset_path", 1);
            define_function!("asset-data_url", "asset_data_url", 1);
            define_function!("asset-url", "asset_url", 1);
            define_function!("image-path", "image_path", 1);
            define_function!("video-path", "video_path", 1);
            define_function!("audio-path", "audio_path", 1);
            define_function!("image-url", "image_url", 1);
            define_function!("video-url", "video_url", 1);
            define_function!("audio-url", "audio_url", 1);
            define_function!("javascript-path", "javascript_path", 1);
            define_function!("stylesheet-path", "stylesheet_path", 1);
            define_function!("font-url", "font_url", 1);
            define_function!("javascript-url", "javascript_url", 1);
            define_function!("stylesheet-url", "stylesheet_url", 1);

            let fs = ImporterFs { registry: importer_registry_mutex };

            options = options.fs(&fs);

            log::info!("Starting grass");
            let grass_result = grass::from_string(input_string, &options);

            match grass_result {
                Ok(css) => {
                    log::trace!("grass css received");
                    log::trace!("grass css length {}", css.len());
                    let mut compile_success = CompileSuccess::new();
                    compile_success.css = css;
                    compile_response.set_success(compile_success);
                },
                Err(err) => {
                    log::error!("grass failed");
                    log::error!("{}", err);
                    compile_response.set_failure( CompileFailure { message: err.to_string(), ..Default::default() } );
                }
            }
        },
        false => {
            log::error!("file input is not implemented")
        }
    }

    let mut outbound_message = OutboundMessage::new();
    outbound_message.set_compile_response(compile_response);
    write_message(outbound_message, id)?;
    log::info!("Compiler request complete");
    Ok(())
}
