use std::{
    collections::HashMap,
    error::Error,
    path::{Path, PathBuf},
};

use clean_path::Clean;
use log::{debug, info, trace, warn};
use proto::sass::{
    inbound_message::{CanonicalizeResponse, ImportResponse},
    outbound_message::{CanonicalizeRequest, ImportRequest},
    OutboundMessage,
};
use url::Url;

use crate::{
    buf_io::{read_message, write_message, MessageType},
    importer_registry::Importer,
};

#[derive(Debug)]
pub struct RemoteImporter {
    id: u32,
    cache: HashMap<String, String>,
    communication_id: u64,
    parent_path: String,
}

impl Importer for RemoteImporter {
    fn canonicalize(&mut self, path: &Path) -> Result<PathBuf, std::io::Error> {
        info!("canonicalize: {:?}", path);

        let path_str = path
            .to_str()
            .ok_or_else(|| std::io::Error::new(std::io::ErrorKind::InvalidInput, "Invalid path"))?
            .to_string();

        if path_str == "stdin" {
            info!("\"stdin\" encountered returning: {:?}", self.parent_path);
            return Ok(PathBuf::from(&self.parent_path));
        }

        let parsed_url = Url::parse(&path_str);
        let absolute = path.is_absolute();

        debug!("path: {:?}, absolute: {}", path, absolute);

        let path = match (parsed_url, absolute) {
            (Ok(x), _) => {
                info!("parsed: {:?}", x);
                debug!("schema: {}", x.scheme());
                match x.scheme() {
                    "file" => {
                        info!("making file absolute");
                        let path_buf = Path::new("/").join(x.path()).clean();
                        info!("absolute path: {:?}", path_buf);
                        let output = Path::new(&format!("file://{}", path_buf.to_str().unwrap()))
                            .to_path_buf();
                        info!("final output path: {:?}", output);
                        output
                    }
                    _ => PathBuf::from(path),
                }
            }
            (Err(_), true) => {
                let path = Path::new(&format!("file://{}", path.to_str().unwrap())).to_path_buf();
                // info!("absolute path encountered returning: {:?}", path);
                PathBuf::from(path)
            }
            (Err(_), false) => {
                warn!("invalid path encountered returning error (invalid path)");
                return Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidInput,
                    "Invalid path",
                ));
            }
        };

        let path_str = path
            .to_str()
            .ok_or_else(|| std::io::Error::new(std::io::ErrorKind::InvalidInput, "Invalid path"))?
            .to_string();

        if let Some(url) = self.cache.get(&path_str) {
            info!("Using cache: {:?}, {:?}", path, url);
            return Ok(PathBuf::from(url));
        }

        let mut response = self
            .send_canonicalize_request(path_str)
            .expect("Could not send canonicalize request");

        let error = response.take_error();
        let url = response.take_url();

        if !error.is_empty() {
            // return empty path
            Ok(PathBuf::new())
        } else if !url.is_empty() {
            let path_buf = PathBuf::from(url.clone());

            if path_buf
                .to_str()
                .ok_or_else(|| {
                    std::io::Error::new(std::io::ErrorKind::InvalidInput, "Invalid path")
                })?
                .starts_with("file://")
            {
                info!("\"file\" encountered caching: {:?}", url);
                let parent = path_buf
                    .parent()
                    .ok_or_else(|| {
                        std::io::Error::new(std::io::ErrorKind::InvalidInput, "Invalid path")
                    })?
                    .to_str()
                    .ok_or_else(|| {
                        std::io::Error::new(std::io::ErrorKind::InvalidInput, "Invalid path")
                    })?
                    .to_string();

                self.cache.insert(url.clone(), parent);
            }
            Ok(path_buf)
        } else {
            Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "No url returned",
            ))
        }
    }

    fn load(&mut self, path: &Path) -> Result<Vec<u8>, std::io::Error> {
        info!("read: {:?}", path);
        let mut response = self
            .send_load_request(path.to_str().unwrap().to_string())
            .expect("Could not send load request");

        let error = response.take_error();
        let content = response.take_success().contents;

        if !error.is_empty() {
            return Err(std::io::Error::new(std::io::ErrorKind::Other, error));
        }

        Ok(content.to_string().into_bytes().to_vec())
    }
}

impl RemoteImporter {
    pub fn new(id: u32, communication_id: u64, path: &str) -> Self {
        Self {
            id,
            cache: HashMap::new(),
            communication_id,
            parent_path: path.to_owned(),
        }
    }

    fn send_canonicalize_request(
        &self,
        path: String,
    ) -> Result<CanonicalizeResponse, Box<dyn Error>> {
        let mut req = CanonicalizeRequest::new();
        req.url = path;
        req.importer_id = self.id;
        req.from_import = true;
        let id = rand::random::<u32>();
        info!("id: {}", id);
        req.id = id;

        let mut msg = OutboundMessage::new();
        msg.set_canonicalize_request(req);

        write_message(msg, self.communication_id)?;
        loop {
            let response = read_message()?;
            info!("got event: {:#?}", response);
            trace!("got event");
            if response.1 == self.communication_id {
                trace!("got event with correct id");
                trace!("a");
                trace!("response: {:?}", &response.0);
                if let MessageType::CanonicalizeResponse(resp) = response.0 {
                    trace!("got event with canonicalize response");
                    return Ok(resp);
                } else {
                    trace!("got event with wrong response type")
                }
            }
        }
    }

    fn send_load_request(&self, path: String) -> Result<ImportResponse, Box<dyn Error>> {
        let mut req = ImportRequest::new();
        req.url = path;
        req.importer_id = self.id;
        let id = rand::random::<u32>();
        req.id = id;
        trace!("id: {}", id);
        let mut msg = OutboundMessage::new();
        msg.set_import_request(req);

        write_message(msg, self.communication_id)?;
        loop {
            let response = read_message()?;
            trace!("got event: {:#?}", response);
            trace!("got event");
            if response.1 == self.communication_id {
                trace!("got event with correct id");
                if let MessageType::ImportResponse(resp) = response.0 {
                    trace!("got event with import response");
                    return Ok(resp);
                }
            }
        }
    }
}
