use std::{path::{Path, PathBuf}, fmt::Debug};

use proto::sass::inbound_message::compile_request::Importer as ImporterType;
use proto::sass::inbound_message::compile_request::importer::Importer as ImportImportType;

use crate::import::{FilesystemImporter, RemoteImporter, HostImporter};
pub trait Importer: Debug {

    fn canonicalize(&mut self, path: &Path) -> std::io::Result<PathBuf>;
    fn load(&mut self, path: &Path) -> std::io::Result<Vec<u8>>;
}

#[derive(Debug)]
pub struct ImporterRegistry {
    importers: Vec<Box<dyn Importer>>,
    communication_id: u64,
    parent_path: String
}

impl ImporterRegistry {
    pub fn new(id: u64, parent_path: String) -> Self {
        Self {
            importers: Vec::new(),
            communication_id: id,
            parent_path
        }
    }
    pub fn register(&mut self, importer: Box<ImporterType>) {
        let importer_type = importer.importer.unwrap();
        match importer_type {
            ImportImportType::FileImporterId(id) => {
                self.register_raw(Box::new(FilesystemImporter::new(id)));
            }
            ImportImportType::ImporterId(id) => {
                self.register_raw(Box::new(RemoteImporter::new(id, self.communication_id, &self.parent_path)));
            }
            ImportImportType::Path(p) => {
                let path = PathBuf::from(p);
                self.register_raw(Box::new(HostImporter::new(path)));
            }
            _ => {
                unreachable!()
            }
        }
        
    }
    pub fn register_raw(&mut self, importer: Box<dyn Importer>) {
        self.importers.push(importer);

    }
    pub fn load(&mut self, path: &Path) -> std::io::Result<Vec<u8>> {
        let mut n = 0;
        log::debug!("Loading path: {:?}", path);
        for importer in &mut self.importers {
            let result = importer.load(path);
            log::debug!("Importer[{}] result ok: {:?}", n, result.is_ok());
            n+=1;
            if result.is_ok() {
                log::debug!("Loaded path: {:?}", path);
                log::debug!("Loaded path data length: {:?}", result.as_ref().unwrap().len());
                return result;
            }
        }
        Err(std::io::Error::new(std::io::ErrorKind::NotFound, "File not found"))
    }
    pub fn canonicalize(&mut self, path: &Path) -> std::io::Result<PathBuf> {
        log::debug!("Canonicalizing path: {:?}", path);
        let mut n = 0;
        for importer in &mut self.importers {
            let result = importer.canonicalize(path);
            log::debug!("Importer[{}] result: {:?}", n, result);
            n+=1;
            if result.is_ok() {
                log::debug!("Canonicalized path: {:?}", result.as_ref().unwrap());
                return result;
            }
        }
        Err(std::io::Error::new(std::io::ErrorKind::NotFound, "File not found"))
    }
    pub fn set_parent_path(&mut self, parent_path: String) {
        self.parent_path = parent_path;
    }
}

